@@ -10,6 +10,15 @@ from tsforest.config import (calendar_features_names,
                              calendar_cyclical_features_names)
 from sklearn.base import TransformerMixin
 
+# available time feature
+AVAILABLE_TIME_FEATURES = ["year", "quarter", "month", "days_in_month",
+                           "year_week", "year_day", "month_day", "week_day",
+                           "hour", "minute", "second", "microsecond", "millisecond"
+                           "nanosecond", "month_progress", "second_cos", "second_sin",
+                           "minute_cos", "minute_sin", "hour_cos", "hour_sin", 
+                           "week_day_cos", "week_day_sin", "year_day_cos", "year_day_sin",
+                           "year_week_cos", "year_week_sin", "month_cos", "month_sin"]
+
 # available methods in pandas.core.window.Rolling as of pandas 0.25.1 
 AVAILABLE_RW_FUNCTIONS = ["count", "sum", "mean", "median", "var", "std", "min", 
                           "max", "corr", "cov", "skew", "kurt", "quantile"]
@@ -26,10 +35,8 @@ class ForecasterBase(object):
     ----------
     model_params : dict
         Dictionary containing the parameters of the specific boosting model.
-    feature_sets: list
-        List of feature sets to be included.
-    exclude_features: list
-        List of features to be excluded from training dataframe.
+    time_features: list
+        Time attributes to include as features.
     categorical_features: dict
         Dict with the name of the categorical feature as keys, and the name
         of the class in 'category_encoders' to be used for encoding as values.
@@ -40,27 +47,26 @@ class ForecasterBase(object):
     trend_models: dict
         Dictionary with ts_uid as key and TrendModel instances as value.
     target_scalers: dict
-        Dictionary with ts_uid as key and TrendModel instances as value.
+        Dictionary with ts_uid as key and sklearn.base.TransformerMixin as
+        value, a trained sklearn.base.TransformerMixin object.
     lags: list
-        List of integer lag values.
+        List of integer lag values to include as features.
     window_sizes: list
-        List of integer window sizes values.
+        List of integer window sizes values to include as features.
     window_functions: list
-        List of string names of the window functions.
+        List of string names of the window functions to include as features.
+    copy: bool
+        If True performs a copy over the dataframes provided, if False performs
+        inplace operations.
     """
-    def __init__(self, model_params=dict(), feature_sets=["calendar", "calendar_cyclical"], 
-                 exclude_features=list(), categorical_features=dict(), calendar_anomaly=list(), 
-                 ts_uid_columns=list(), trend_models=dict(), target_scalers=dict(), 
-                 lags=None, window_sizes=None, window_functions=None):
-
+    def __init__(self, model_params=dict(), time_features=[], exclude_features=[], 
+                 categorical_features=dict(), calendar_anomaly=list(), ts_uid_columns=list(), 
+                 trend_models=dict(), target_scalers=dict(), lags=list(), window_sizes=list(), 
+                 window_functions=list(), copy=False):
         self.model = None
         self.model_params = model_params
-        self.feature_sets = feature_sets.copy()
-        if lags is not None and "lag" not in feature_sets:
-            self.feature_sets.append("lag")
-        if (window_sizes is not None and window_functions is not None) and "rw" not in feature_sets:
-            self.feature_sets.append("rw")    
-        self.exclude_features = ["ds", "y", "y_raw", "weight", "fold_column",
+        self.time_features = time_features
+        self.exclude_features = ["ds", "y", "y_raw", "weight", "fold_column", "trend",
                                  "zero_response", "calendar_anomaly"] + exclude_features
         self.categorical_features = categorical_features.copy()
         self.calendar_anomaly = calendar_anomaly
@@ -73,6 +79,7 @@ class ForecasterBase(object):
         self.lags = lags
         self.window_sizes = window_sizes
         self.window_functions = window_functions
+        self.copy = copy
         self._features_already_prepared = False
         self._validate_inputs()
     
@@ -89,12 +96,15 @@ class ForecasterBase(object):
         """
         if not isinstance(self.model_params, dict):
             raise TypeError("Parameter 'model_params' should be of type 'dict'.")
-
-        if not isinstance(self.feature_sets, list):
-            raise TypeError("Parameter 'feature_sets' should be of type 'list'.")
+    
+        if not isinstance(self.time_features, list):
+            raise TypeError("Parameter 'time_features' should be of type 'list'.")
         else:
-            if any([x not in ["calendar", "calendar_cyclical", "lag", "rw"] for x in self.feature_sets]):
-                raise ValueError("Values in 'feature_sets' should be any of: ['calendar', 'calendar_cyclical', 'lag', 'rw'].")
+            if any([feature not in AVAILABLE_TIME_FEATURES for feature in self.time_features]):
+                raise ValueError(f"Values in 'time_features' should by any of: {AVAILABLE_TIME_FEATURES}")
+        
+        if not isinstance(self.exclude_features, list):
+            raise TypeError("Parameter 'exclude_features' should be of type 'list'.")
         
         if not isinstance(self.categorical_features, dict):
             raise TypeError("Parameter 'categorical_features' should be of type 'dict'.")
@@ -118,32 +128,29 @@ class ForecasterBase(object):
         elif not all([isinstance(scaler, TransformerMixin) for scaler in self.target_scalers.values()]):
             raise ValueError("Values in 'target_scalers' shoud be instances of 'TransformerMixin'.")
 
-        if self.lags is not None:
-            if not isinstance(self.lags, list):
-                raise TypeError("Parameter 'lags' should be of type 'list'.")
-            else:
-                if any([type(x)!=int for x in self.lags]):
-                    raise ValueError("Values in 'lags' should be integers.")
-                elif any([x<1 for x in self.lags]):
-                    raise ValueError("Values in 'lags' should be integers greater or equal to 1.")
+        if not isinstance(self.lags, list):
+            raise TypeError("Parameter 'lags' should be of type 'list'.")
+        else:
+            if any([type(x)!=int for x in self.lags]):
+                raise ValueError("Values in 'lags' should be integers.")
+            elif any([x<1 for x in self.lags]):
+                raise ValueError("Values in 'lags' should be integers greater or equal to 1.")
         
-        if self.window_sizes is not None:
-            if not isinstance(self.window_sizes, list):
-                raise TypeError("Parameter 'window_sizes' should be of type 'list'.")
-            else:
-                if any([type(x)!=int for x in self.window_sizes]):
-                    raise ValueError("Values in 'window_sizes' should be integers.")
-                elif any([x<1 for x in self.window_sizes]):
-                    raise ValueError("Values in 'window_sizes' should be integers greater or equal to 1.")
+        if not isinstance(self.window_sizes, list):
+            raise TypeError("Parameter 'window_sizes' should be of type 'list'.")
+        else:
+            if any([type(x)!=int for x in self.window_sizes]):
+                raise ValueError("Values in 'window_sizes' should be integers.")
+            elif any([x<1 for x in self.window_sizes]):
+                raise ValueError("Values in 'window_sizes' should be integers greater or equal to 1.")
         
-        if self.window_functions is not None:
-            if not isinstance(self.window_functions, list):
-                raise TypeError("Parameter 'window_functions' should be of type list.")
-            else:
-                if any([type(x)!=str for x in self.window_functions]):
-                    raise ValueError("Values in 'window_functions' should be string names.")
-                elif any([x not in AVAILABLE_RW_FUNCTIONS for x in self.window_functions]):
-                    raise ValueError(f"Values in 'window_functions' should be any of: {AVAILABLE_RW_FUNCTIONS}.")
+        if not isinstance(self.window_functions, list):
+            raise TypeError("Parameter 'window_functions' should be of type list.")
+        else:
+            if any([type(x)!=str for x in self.window_functions]):
+                raise ValueError("Values in 'window_functions' should be string names.")
+            elif any([x not in AVAILABLE_RW_FUNCTIONS for x in self.window_functions]):
+                raise ValueError(f"Values in 'window_functions' should be any of: {AVAILABLE_RW_FUNCTIONS}.")
     
     def _validate_input_data(self, train_data, valid_index):
         if not isinstance(train_data, pd.DataFrame):
@@ -158,17 +165,20 @@ class ForecasterBase(object):
 
         if not set(self.ts_uid_columns) <= set(train_data.columns):
             raise ValueError(f"Parameter 'train_data' has missing ts_uid_columns: {set(self.ts_uid_columns)-set(train_data.columns)}.")
+        
+        if {"internal_ts_uid", "_internal_ts_uid"} in set(train_data.columns):
+            raise ValueError("Columns 'internal_ts_uid' and '_internal_ts_uid' is reserved for internal usage and can not be in 'train_data' dataframe.")
     
     def _validate_predict_data(self, predict_data):
         if not isinstance(predict_data, pd.DataFrame):
             raise TypeError("Parameter 'predict_data' should be of type pandas.DataFrame.")
-        elif not (set(self.train_data.columns) - set(predict_data.columns) == {"y", "y_raw"}):
+        elif not (set(self.raw_train_columns) - set(predict_data.columns) == {"y"}):
             raise ValueError("'predict_data' shoud have the same columns as 'train_data' except for 'y'.")
     
     def _validate_evaluate_data(self, eval_data, metric):
         if not isinstance(eval_data, pd.DataFrame):
             raise TypeError("'eval_data' should be of type pandas.DataFrame.")
-        elif not (set(eval_data.columns) <= set(self.train_data.columns)):
+        elif not (set(self.raw_train_columns) == set(eval_data.columns)):
             raise ValueError("'eval_data' should have the same columns as 'train_data'.")
 
         if not isinstance(metric, str):
@@ -176,48 +186,25 @@ class ForecasterBase(object):
         elif metric not in AVAILABLE_METRICS:
             raise ValueError(f"'metric' should be any of these: {AVAILABLE_METRICS}")
     
-    def _encode_categorical_features(self, train_features, categorical_features, ts_uid_columns):
+    def _encode_categorical_features(self, train_features):
         categorical_encoders = dict()
-        for feature,encoding in categorical_features.items():
+        for feature,encoding in self.categorical_features.items():
             if encoding == "default": 
                 if not np.issubdtype(train_features[feature].dtype, np.number):
                     encoding = "OrdinalEncoder"
                 else: continue
             encoder_class = getattr(ce, encoding)
             encoder = encoder_class(cols=[feature])
-            if feature in ts_uid_columns:
-                encoder.fit(train_features.loc[:, [feature]], train_features.loc[:, "y_raw"].values)
-            else:
-                encoder.fit(train_features.loc[:, [feature]], train_features.loc[:, "y"].values)
+            encoder.fit(train_features.loc[:, [feature]], train_features.loc[:, "y_raw"].values)
             transformed = encoder.transform(train_features.loc[:, [feature]])
-            del train_features[feature]
-            train_features[transformed.columns] = transformed 
+            if feature in self.ts_uid_columns:
+                train_features["_"+feature] = transformed.values
+                self.exclude_features.append(feature)
+            else:
+                del train_features[feature]
+                train_features[feature] = transformed.values
             categorical_encoders[feature] = encoder
         return train_features,categorical_encoders
-      
-    def _prepare_train_features(self, train_data):
-        """
-        Parameters
-        ----------
-        train_data : pandas.DataFrame
-            Dataframe with at least columns 'ds' and 'y'.
-        """
-        train_features = compute_train_features(data=train_data,
-                                                include_features=self.feature_sets,
-                                                lags=self.lags,
-                                                window_sizes=self.window_sizes,
-                                                window_functions=self.window_functions,
-                                                ignore_const_cols=True)
-        if "zero_response" in train_features.columns:
-            train_features = train_features.query("zero_response != 1")
-        if "calendar_anomaly" in train_features.columns:
-            assert len(self.calendar_anomaly) != 0, \
-                "'calendar_anomaly' column found, but names of affected features were not provided."
-            assert set(self.calendar_anomaly) <= set(train_features.columns), \
-                f"Calendar anomaly affected columns: {set(self.calendar_anomaly)-set(train_features.columns)} are not present in 'train_features'."
-            idx = train_features.query("calendar_anomaly == 1").index
-            train_features.loc[idx, self.calendar_anomaly] = np.nan
-        return train_features
     
     def _prepare_predict_features(self, predict_data):
         """
@@ -233,7 +220,8 @@ class ForecasterBase(object):
             the trained model.
         """
         predict_features = compute_predict_features(data=predict_data,
-                                                    include_features=self.feature_sets,
+                                                    ts_uid_columns=self.ts_uid_columns,
+                                                    time_features=self.time_features,
                                                     lags=self.lags,
                                                     window_sizes=self.window_sizes,
                                                     window_functions=self.window_functions,
@@ -246,56 +234,69 @@ class ForecasterBase(object):
                 f"Calendar anomaly affected columns: {set(self.calendar_anomaly)-set(self.train_features.columns)} are not present in 'train_features'."
             idx = predict_features.query("calendar_anomaly == 1").index
             predict_features.loc[idx, self.calendar_anomaly] = np.nan
-        if len(self.categorical_features) > 0:
+        if len(self.categorical_encoders) > 0:
             for feature,encoder in self.categorical_encoders.items():
                 transformed = encoder.transform(predict_features.loc[:, [feature]])
-                del predict_features[feature]
-                predict_features[transformed.columns] = transformed 
-
+                if feature in self.ts_uid_columns:
+                    predict_features["_"+feature] = transformed.values
+                else:
+                    del predict_features[feature]
+                    predict_features[feature] = transformed.values
         features_to_keep = [feature for feature in predict_features.columns if feature in self.raw_features]
         return predict_features.loc[:, features_to_keep]
     
-    def _prepare_target(self, data, trend_model, target_scaler):
+    def prepare_target(self, train_data):
         """
         Prepares the target variable
 
         Parameters
         ----------
-        data: pd.DataFrame
+        train_data: pd.DataFrame
             Dataframe containing the columns 'ds' and 'y'.
-        trend_model: TrendModel
-            Trained trend estimator model.
-        target_scaler: sklearn.base.TransformerMixin
-            Trained sklearn.base.TransformerMixin scaler.
         """
-        data = data.copy()
-        data.y = data.y.astype(float)
-        if trend_model is not None:
-            trend_dataframe = trend_model.predict(data.loc[:, ["ds"]]) 
-            data.loc[:, "y"] -= trend_dataframe.trend.values
-        if target_scaler is not None:
-            data.loc[:, "y"] = target_scaler.transform(data.y.values.reshape(-1,1)).ravel()
-        return data.y.values
+        train_data["y_raw"] = train_data["y"].copy()
+        ts_uid_values = train_data.loc[:, self.ts_uid_columns].drop_duplicates()
+        all_dataframes = list()
+        for _,row in ts_uid_values.iterrows():
+            key = tuple([item for _,item in row.iteritems()])
+            query_string = " & ".join([f"{col_name}=={value}" for col_name,value in row.iteritems()])
+            train_data_slice = train_data.query(query_string).copy()
 
-    def prepare_train_features(self, train_data, trend_model, target_scaler, sort_by):
+            trend_model = self.trend_models[key] if len(self.trend_models) > 0 else None
+            target_scaler = self.target_scalers[key] if len(self.target_scalers) > 0 else None
+            if trend_model is not None:
+                trend_dataframe = trend_model.predict(train_data_slice.loc[:, ["ds"]]) 
+                train_data_slice["trend"] = trend_dataframe.trend.values
+                train_data_slice.loc[:, "y"] -= trend_dataframe.trend.values
+            if target_scaler is not None:
+                train_data_slice.loc[:, "y"] = target_scaler.transform(train_data_slice.y.values.reshape(-1,1)).ravel()
+            all_dataframes.append(train_data_slice)
+        
+        return pd.concat(all_dataframes)
+
+    def prepare_train_features(self, train_data):
         """
         Parameters
         ----------
         train_data : pandas.DataFrame
             Dataframe with at least columns 'ds' and 'y'.
-        trend_model: TrendModel
-            Trained trend estimator model.
-        target_scaler: sklearn.base.TransformerMixin
-            Trained sklearn.base.TransformerMixin scaler.
-        sort_by: list
-            List of column names to sort 'train_data'
         """
-        train_data = train_data.sort_values(sort_by, axis=0)
-        y_target = self._prepare_target(train_data, trend_model, target_scaler)
-        train_data["y_raw"] = train_data.pop("y").values
-        train_data["y"] = y_target
-        train_features = self._prepare_train_features(train_data)
-        train_features.set_index(train_data.index, inplace=True)
+        train_features = compute_train_features(data=train_data,
+                                                ts_uid_columns=self.ts_uid_columns,
+                                                time_features=self.time_features,
+                                                lags=self.lags,
+                                                window_sizes=self.window_sizes,
+                                                window_functions=self.window_functions,
+                                                ignore_const_cols=True)
+        if "zero_response" in train_features.columns:
+            train_features = train_features.query("zero_response != 1")
+        if "calendar_anomaly" in train_features.columns:
+            assert len(self.calendar_anomaly) != 0, \
+                "'calendar_anomaly' column found, but names of affected features were not provided."
+            assert set(self.calendar_anomaly) <= set(train_features.columns), \
+                f"Calendar anomaly affected columns: {set(self.calendar_anomaly)-set(train_features.columns)} are not present in 'train_features'."
+            idx = train_features.query("calendar_anomaly == 1").index
+            train_features.loc[idx, self.calendar_anomaly] = np.nan
         return train_features
     
     def prepare_features(self, train_data, valid_index=pd.Int64Index([])):
@@ -308,44 +309,41 @@ class ForecasterBase(object):
             Array with indexes from train_data to be used for validation.
         """
         self._validate_input_data(train_data, valid_index)
-        # adds a dummy 'ts_uid' column in case of single ts data
-        if len(self.ts_uid_columns) == 0: 
-            train_data["ts_uid"] = 0
-            ts_uid_columns = ["ts_uid"]
+        self.raw_train_columns = train_data.columns
+        if self.copy:
+            train_data = train_data.copy(deep=True)
+        if len(self.ts_uid_columns) == 0:
+            # adds a dummy '_internal_ts_uid' column in case of single ts data
+            train_data["_internal_ts_uid"] = 0
+            self.ts_uid_columns = ["_internal_ts_uid"]
+        train_data = train_data.sort_values(self.ts_uid_columns+["ds"], axis=0)
+        
+        if len(self.trend_models) > 0 or len(self.target_scalers) > 0:
+            train_data = self.prepare_target(train_data)
         else:
-            ts_uid_columns = self.ts_uid_columns
+            train_data["y_raw"] = train_data["y"].values
+        train_features = self.prepare_train_features(train_data)
+        # needed to keep track of the rows used in valid_index
+        train_features.set_index(train_data.index, inplace=True)
 
-        if (len(self.trend_models) == 0 and
-            len(self.target_scalers) == 0 and 
-            {"lag", "rw"} & set(self.feature_sets) == set()):
-              train_features = self.prepare_train_features(train_data, None, None, sort_by=ts_uid_columns+["ds"])
-        else:
-            all_train_features = list()
-            ts_uid_values = train_data.loc[:, ts_uid_columns].drop_duplicates()
-            for _,row in ts_uid_values.iterrows():
-                key = tuple([item for _,item in row.iteritems()])
-                query_string = " & ".join([f"{col_name}=={value}" for col_name,value in row.iteritems()])
-                train_data_chunk = train_data.query(query_string)
-                trend_model = self.trend_models[key] if len(self.trend_models) > 0 else None
-                target_scaler = self.target_scalers[key] if len(self.target_scalers) > 0 else None
-                train_features = self.prepare_train_features(train_data_chunk, trend_model, target_scaler, sort_by=["ds"])
-                all_train_features.append(train_features)         
-            train_features = pd.concat(all_train_features)
-
-        self.train_data = train_features.loc[:, list(train_data.columns) + ["y_raw"]]
         if len(valid_index) > 0:
             valid_features = train_features.loc[valid_index, :]
             train_features = train_features.drop(valid_index, axis=0)
  
         # performs the encoding of categorical features
         if len(self.categorical_features) > 0:
-            train_features,categorical_encoders = self._encode_categorical_features(train_features, self.categorical_features, ts_uid_columns)
+            train_features,categorical_encoders = self._encode_categorical_features(train_features)
             if len(valid_index) > 0:
                 for feature,encoder in categorical_encoders.items():
                     transformed = encoder.transform(valid_features.loc[:, [feature]])
-                    del valid_features[feature]
-                    valid_features[transformed.columns] = transformed 
+                    if feature in self.ts_uid_columns:
+                        valid_features["_"+feature] = transformed.values
+                    else:
+                        del train_features[feature]
+                        valid_features[feature] = transformed.values
             self.categorical_encoders = categorical_encoders
+        else:
+            self.categorical_encoders = dict()
         # categorical features to be encoded by the tree/boosting model
         _categorical_features = [feature for feature,encoder in self.categorical_features.items() 
                                  if encoder == "default"]
@@ -354,6 +352,7 @@ class ForecasterBase(object):
         self.raw_features = train_features.columns
         self.input_features = [feature for feature in train_features.columns
                                if feature not in self.exclude_features]
+        self.train_data = train_data
         self.train_features = train_features
         self.valid_features = valid_features if len(valid_index) > 0 else None
         self._features_already_prepared = True
@@ -384,99 +383,102 @@ class ForecasterBase(object):
         self.model.fit(**kwargs)
         self.best_iteration = self.model.best_iteration
 
-    def predict(self, predict_data):
+    def predict(self, predict_data, recursive=False):
         """
         Parameters
         ----------
         predict_data: pandas.DataFrame
             Datafame containing the features for the prediction period.
             Contains the same columns as 'train_data' except for 'y'.
+        recursive: boolean
+            If True, perform recursive one-step-ahead predicion for the
+            lag and/or rolling window features.
         Returns
         ----------
         prediction_dataframe: pandas.DataFrame
             Dataframe containing the dates 'ds' and predictions 'y_pred'.
         """
-        # adds a dummy 'ts_uid' column in case of single ts data
-        if len(self.ts_uid_columns) == 0: 
-            predict_data["ts_uid"] = 0
-            ts_uid_columns = ["ts_uid"]
-        else:
-            ts_uid_columns = self.ts_uid_columns
-        self._validate_predict_data(predict_data) 
+        self._validate_predict_data(predict_data)
+        if self.copy:
+            predict_data = predict_data.copy(deep=True)
+        if set(self.ts_uid_columns) == {"_internal_ts_uid"}:
+            # adds a dummy '_internal_ts_uid' column in case of single ts data
+            predict_data["_internal_ts_uid"] = 0
 
-        if (len(self.trend_models) == 0 and
-            len(self.target_scalers) == 0 and
-            {"lag", "rw"} & set(self.feature_sets) == set()):
-            predict_features = self._prepare_predict_features(predict_data)
+        predict_features = self._prepare_predict_features(predict_data)
+        if not recursive:
             prediction = self.model.predict(predict_features)
-            if "zero_response" in predict_features.columns:
-                zero_response_mask = predict_features["zero_response"]==1
-                prediction[zero_response_mask] = 0
-            self.predict_features = predict_features
             prediction_dataframe = (predict_data
                                     .loc[:, ["ds"]+self.ts_uid_columns]
                                     .assign(y_pred = prediction))
         else:
-            ts_uid_values = predict_data.loc[:, ts_uid_columns].drop_duplicates()
-            all_predict_features = list()
-            all_prediction_dataframes = list()
+            predict_features.sort_values(self.ts_uid_columns+["ds"], axis=0, inplace=True)
+            _prediction_dataframe = self.recursive_predict(predict_features)
+            prediction_dataframe = pd.merge(predict_data.loc[:, ["ds"]+self.ts_uid_columns],
+                                            _prediction_dataframe, 
+                                            how="left", on=["ds"]+self.ts_uid_columns)
+        
+        if (len(self.trend_models) > 0 or
+            len(self.target_scalers) > 0):
+            ts_uid_values = prediction_dataframe.loc[:, self.ts_uid_columns].drop_duplicates()
             for _,row in ts_uid_values.iterrows():
                 key = tuple([item for _,item in row.iteritems()])
                 query_string = " & ".join([f"{col_name}=={value}" for col_name,value in row.iteritems()])
-                predict_data_chunk = predict_data.query(query_string)
-                predict_features = self._prepare_predict_features(predict_data_chunk)
-                if "lag" in self.feature_sets or "rw" in self.feature_sets:
-                    y_past = self.train_data.query(query_string).y.values
-                    prediction = self._predict(self.model, predict_features, y_past)
-                else:
-                    prediction = self.model.predict(predict_features)
+                slice_idx = prediction_dataframe.query(query_string).index
                 
                 if len(self.target_scalers) > 0:
                     target_scaler = self.target_scalers[key]
-                    prediction = target_scaler.inverse_transform(prediction.reshape(-1,1)).ravel()
+                    target_scaled = prediction_dataframe.loc[slice_idx,"y_pred"].values.reshape(-1,1)
+                    target_unscaled = target_scaler.inverse_transform(target_scaled).ravel()
+                    prediction_dataframe.loc[slice_idx, "y_pred"] = target_unscaled
                 if len(self.trend_models) > 0:
                     trend_model = self.trend_models[key]
-                    trend_dataframe = trend_model.predict(predict_data_chunk.loc[:, ["ds"]])
-                    prediction += trend_dataframe.trend.values
-                if "zero_response" in predict_features.columns:
-                    zero_response_mask = predict_features["zero_response"]==1
-                    prediction[zero_response_mask] = 0
-                
-                _prediction_dataframe = (predict_data_chunk
-                                         .loc[:, ["ds"]+self.ts_uid_columns]
-                                         .assign(y_pred = prediction))
-                all_predict_features.append(predict_features)
-                all_prediction_dataframes.append(_prediction_dataframe)
-            self.predict_features = pd.concat(all_predict_features).reset_index(drop=True)
-            prediction_dataframe = pd.concat(all_prediction_dataframes).reset_index(drop=True)
+                    trend_dataframe = trend_model.predict(prediction_dataframe.loc[slice_idx, ["ds"]])
+                    prediction_dataframe.loc[slice_idx, "y_pred"] += trend_dataframe.trend.values
+
+        if "zero_response" in predict_features.columns:
+            zero_response_idx = predict_features.query("zero_response == 1").index
+            prediction_dataframe.loc[zero_response_idx, "y_pred"] = 0
+
+        self.predict_features = predict_features
+        return prediction_dataframe
+
+    def recursive_predict(self, predict_features):
+        max_offset = max(0 if len(self.lags)==0 else max(self.lags), \
+                         0 if len(self.window_sizes)==0 else max(self.window_sizes))
+        min_date = self.train_data.ds.max() - pd.DateOffset(max_offset+1)
+        train_temp = (self.train_data
+                      .loc[:, self.ts_uid_columns+["ds","y"]]
+                      .query("ds >= @min_date")
+                      .copy(deep=True))
+
+        for time_step in np.sort(predict_features.ds.unique()):
+            slice_idx = predict_features.query("ds == @time_step").index
+
+            for lag in self.lags: 
+                lag_values = train_temp.groupby(self.ts_uid_columns)["y"].apply(lambda x: x.iloc[-lag])
+                predict_features.loc[slice_idx, f"lag_{lag}"] = lag_values.values
+            
+            for window_func in self.window_functions:
+                for window in self.window_sizes:
+                    rw_values = (train_temp.groupby(self.ts_uid_columns)["y"]
+                                .apply(lambda x: getattr(np, window_func)(x.iloc[-window:])))
+                    predict_features.loc[slice_idx, f"{window_func}_{window}"] = rw_values.values
+            
+            _prediction = self.model.predict(predict_features.loc[slice_idx,:])
+            _prediction_dataframe = (predict_features.loc[slice_idx, ["ds"]+self.ts_uid_columns]
+                                     .assign(y = _prediction))
+
+            train_temp = pd.concat([train_temp, _prediction_dataframe], axis=0, ignore_index=True)
+
+        prediction_dataframe = pd.merge(predict_features.loc[:, ["ds"]+self.ts_uid_columns], 
+                                        train_temp, 
+                                        how="left", 
+                                        on=["ds"]+self.ts_uid_columns)
+        prediction_dataframe.rename({"y":"y_pred"}, axis=1, inplace=True)
         return prediction_dataframe 
 
-    def _predict(self, model, predict_features, y_past):
-        """
-        Parameters
-        ----------
-        model: BaseRegressor
-            Instance model of tsforest.forest module.
-        predict_features: pandas.DataFrame
-            Datafame containing the features for the prediction period.
-        y_past: np.ndarray
-            Array with the values of 'y' previous to the prediction period.
-        """
-        y = y_past.tolist()
-        for idx in predict_features.index:
-            if "lag" in self.feature_sets:
-                for lag in self.lags:
-                    predict_features.loc[idx, f"lag_{lag}"] = y[-lag]
-            if "rw" in self.feature_sets:
-                for window_func in self.window_functions:
-                    for window in self.window_sizes:
-                        predict_features.loc[idx, f"{window_func}_{window}"] = getattr(np, window_func)(y[-window:])
-            y_pred = model.predict(predict_features.loc[[idx], self.input_features])
-            y.append(y_pred)
-        n_predictions = predict_features.shape[0]
-        return np.asarray(y[-n_predictions:])
-
-    def evaluate(self, eval_data, metric="rmse"):
+    def evaluate(self, eval_data, metric="rmse", recursive=False):
         """
         Parameters
         ----------
@@ -492,7 +494,7 @@ class ForecasterBase(object):
         self._validate_evaluate_data(eval_data, metric)
         eval_data = eval_data.copy()
         y_real = eval_data.pop("y")
-        y_pred = self.predict(eval_data)["y_pred"].values
+        y_pred = self.predict(eval_data, recursive)["y_pred"].values
         error_func = getattr(metrics, f"compute_{metric}")
         error = error_func(y_real, y_pred)
         return error
